#include"cache.h"
#include"constants.h"

packet_cache_t* packet_cache_init(int size)
{
    int i;
    packet_cache_t* ret = malloc(sizeof(packet_cache_t));
    ret->cache          = (cache_entry_t**)malloc(sizeof(cache_entry_t*)*size);

    for(i=0; i < size; i++)
        ret->cache[i] = NULL;

    ret->size   = size;
    ret->index  = 0;
    return ret;
}

cache_entry_t* cache_entry_init(int time, queue_t* queue,
                                const u_char* packet, unsigned int len, int id)
{
    cache_entry_t* entry = (cache_entry_t*)malloc(sizeof(cache_entry_t));

    entry->time     = time;
    entry->queue    = queue;
    entry->packet   = g_memdup(packet, len);
    entry->len      = len;
    entry->id       = id;

    return entry;
}

void entry_free(cache_entry_t* entry)
{
    g_free(entry->packet);
    free(entry);
}

void packet_cache_delete(packet_cache_t* cache)
{
    // Do not free nodes here, just packets 
    // Nodes are released when the trees are
    int i;
    for(i=0; i < cache->size; i++)
    {
	    cache_entry_t* entry = cache->cache[i];
	    if(entry != NULL)
            entry_free(cache->cache[i]);
    }
}

void packet_cache_insert(packet_cache_t* cache, queue_t* queue,
                         const u_char* packet, unsigned int len, int id)
{
    if(cache->cache[cache->index] != NULL)
        entry_free(cache->cache[cache->index]);

    cache->cache[cache->index] = cache_entry_init(simtime, queue, packet, len,
                                                  id);
    cache->index++;
    if(cache->index == cache->size)
        cache->index = 0;
}

void packet_cache_free(packet_cache_t* cache, queue_t* queue)
{
    int i;
    for(i=0; i < cache->size; i++)
    {
        cache_entry_t* entry = cache->cache[i];
        if(entry != NULL && entry->queue == queue)
        {
            entry_free(entry);
            cache->cache[i] = NULL;
        }
    }
}

static int packets_with_queue(packet_cache_t* cache, queue_t* queue)
{
    int i;
    int count=0;
    for(i=0; i < cache->size; i++)
    {
        cache_entry_t* entry = cache->cache[i];
        if(entry != NULL && entry->queue == queue)
            ++count;
    }

    return count;
}

png_t* packet_cache_png_find(packet_cache_t* cache, queue_t* queue)
{
    int i;

    // We only want to sample around P_CACHE_MAX_SAMPLE packets.
    // Therefore, we use a sampling probability to down sample packets
    // associated with a particular queue if we're getting more than
    // that.  That's (2) sampling probabilities now.  The first,
    // SAMPLE_PROBABILITY, is a packet's probability of ending in the
    // packet cache.  P_CACHE_MAX_SAMPLE is the maximum number of
    // packets that are sampled for each queue.  Note that the actual
    // probability is high unless there are MANY packets associated
    // with the queue.
    int count = packets_with_queue(cache, queue);
    double prob = 1.0;
    if(count > P_CACHE_MAX_SAMPLE)
        prob = P_CACHE_MAX_SAMPLE*1.0/count;

    // Perhaps should be replaced by g_hsh_table_new_full
    GHashTable* pngs = g_hash_table_new(png_hash_fn, png_hash_cmp);
    png_t* ret;

    count = 0;
    for(i=0; i < cache->size; i++)
    {
        cache_entry_t* entry = cache->cache[i];
        if(entry != NULL && entry->queue == queue)
        {
            // Ignore the packet so that we sample approximately
            // P_CACHE_MAX_SAMPLE packets
            if(drand48() < prob )
            {
                ++count;
                GLIB_packet2pngs(pngs, entry->packet, entry->len);
            }
        }
    }

    if(OPTIMAL)
        ret = GLIB_png_find_optimal(pngs, count);
    else
        ret = GLIB_png_find_sufficient(pngs, count);

    GLIB_destroy_pngs(pngs);
    g_hash_table_destroy(pngs);
    return ret;
}


void packet_cache_stats(FILE* flog, packet_cache_t* cache, queue_t* queue)
{
    int i;
    int count = 0;
    double average = 0.0;
    assert(queue != NULL);      // Need to change this later

    //printf("Size: %d\n", cache->index);
    for(i=0; i < cache->size; i++)
    {
        cache_entry_t* entry = cache->cache[i];
        if(entry != NULL && entry->queue == queue) 
        {
            count++;
            average += cache->cache[i]->time;
        }
    }

    //fprintf(flog, "Queue: %d\t%d\t%f\n", queue, count, average/count);
}

void packet_cache_show_pngs_for_offset(packet_cache_t* cache, queue_t* queue, int offset)
{
    int i, count=0;

    GHashTable* pngs = g_hash_table_new(png_hash_fn, png_hash_cmp);

    for(i=0; i < cache->size; i++)
    {
        cache_entry_t* entry = cache->cache[i];
        if(entry != NULL && entry->queue == queue)
        {
            ++count;
            GLIB_packet2pngs(pngs, entry->packet, entry->len);
        }
    }

    void show_for_offset(gpointer key, gpointer value, gpointer data) {
        png_t* png  = (png_t*)key;
        int count   = GPOINTER_TO_INT(value);

        if(png->offset == offset) {
            png_print(flog, png);
            fprintf(flog, " :-> %d\n", count);
        }
    }

    g_hash_table_foreach(pngs, show_for_offset, NULL); 

    GLIB_destroy_pngs(pngs);
    g_hash_table_destroy(pngs);
}

