#include"config.h"
#include"constants.h"

#include<glib.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>

static char* read_next(int argc, char **argv, int *i, const char *param,
                       bool optional)
{
    if (*i+1 < argc && argv[*i+1][0] != '-')
        return argv[++(*i)];
    else if (!optional) {
        printf("%s requires an argument\n", param);
        exit(1);
    }
    else
        return NULL;
}

static void print_version()
{
    const char* IHV_COPYRIGHT_STRING =
"Copyright (C) 2006-2007  Hajime Inoue\n"
"              2007       Dana Jansens\n\n"
"This is free software; see the source for copying conditions.  There is NO\n"
"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n";

    printf("adhc version %s\n", VERSION);
    printf("%s\n", IHV_COPYRIGHT_STRING);
}

static void print_help()
{
    const char* usage =
"Usage: netadhict [options] file...\n"
"Options:\n"
"  -h, --help                   Display this message\n"
"  -r, --noheaders [size]       The number of bytes at the start of each packet\n"
"                                 to ignore\n"
"      --ip                     Ignore all packets but those with ethertype IPv4\n"
"  -I [ethN]                    Read the specified interface instead of files\n"
"                                 for input\n"
"                                 (Defaults to \"eth0\")\n"
"  -i, --interactive            Start program in command-line mode\n"
"      --noports                Ignore ports in UDP and TCP packets\n"
"                                 (Must be used with --ip)\n"
"  -V, --version                Print version information\n"
"      --verbose                Print out extra information in tree and log files\n"
"  -f, --file                   Use files as tick markers\n"
"\nDatabase Output Options:\n"
"      --db [database]          The name of the database to use for output\n"
"                                 (Defaults to \"netadhict\")\n"
"      --db-host [host]         Optionally specify the host to use to connect to\n"
"                                 (Defaults to \"localhost\")\n"
"      --db-port [port]         Optionally specify the port to connect to\n"
"      --db-user [user]         Optionally specify the database user name\n"
"                                 (Defaults to \"root\")\n"
"      --db-password [passwd]   Optionally specify the database password\n"
"                                 (Defaults to none)\n"
"      --db-tree [tree]         Optionally specify the name of the tree to\n"
"                                 output to (Defaults to \"default\")\n"
"\nDot-file Output Options:\n"
"      --dot [file path]        The name of the file to use for dot output.\n"
"\n\nExamples:\n"
"      netadhict --db\n"
"          Read eth0 and write to the database.\n"
"      netadhict -I eth1 --dot out.dot\n"
"          Read eth1 and write a dotfile as output.\n"
"      netadhict --db -- packets0.pcap packets1.pcap packets2.pcap\n"
"          Read tcpdump files as input and write to the database.\n"
"      netadhict --db netadhict packets0.pcap packets1.pcap packets2.pcap\n"
"          Read tcpdump files as input and write to the database.\n"
"      netadhict --dot out.dot packets0.pcap packets1.pcap packets2.pcap\n"
"          Read tcpdump files as input and write a dotfile as output.\n"
"\n";

    printf("%s", usage);
}

config_t* config_load(int argc, char **argv)
{
    config_t *c = g_new0(config_t, 1);
    gint i;
    bool force_interface = false;

    /* defaults */
    c->interface = "eth0";
    c->database_tree = "default";
    c->database_host = "localhost";
    c->database_user = "root";
    c->database_name = "netadhict";

    for(i = 1; i < argc; ++i) {
        if(strcmp(argv[i], "--help") == 0 ||
           strcmp(argv[i], "-h") == 0) {
            print_help();
            exit(EXIT_SUCCESS);
        }
        else if(strcmp(argv[i], "--version") == 0 ||
                strcmp(argv[i], "-V") == 0) {
            print_version();
            exit(EXIT_SUCCESS);
        }

        else if(strcmp(argv[i], "--ip") == 0)
            c->ip_only = true;

        else if(strcmp(argv[i], "-I") == 0) {
            c->interface = read_next(argc, argv, &i, "-I", false);
            force_interface = true;
        }
        else if(strcmp(argv[i], "--interactive") == 0 ||
                 strcmp(argv[i], "-i") == 0) {
            c->interactive = true;
        }
        else if(strcmp(argv[i], "--noports") == 0) {
            c->ignore_port = true;
        }
        else if(strcmp(argv[i], "--verbose") == 0) {
            c->verbose = true;
        }
        else if(strcmp(argv[i], "--file") == 0 ||
                 strcmp(argv[i], "-f") == 0) {
            c->by_file = true;
        }

        else if(strcmp(argv[i], "--db-host") == 0)
            c->database_host = read_next(argc, argv, &i, "--db-host", false);
        else if(strcmp(argv[i], "--db-port") == 0)
            c->database_port = atoi(read_next(argc, argv, &i, "--db-port",
                                              false));
        else if(strcmp(argv[i], "--db-user") == 0)
            c->database_user = read_next(argc, argv, &i, "--db-user", false);
        else if(strcmp(argv[i], "--db-password") == 0)
            c->database_pass = read_next(argc, argv, &i, "--db-password",
                                         false);
        else if(strcmp(argv[i], "--db") == 0) {
            char *n;
            c->database_output = true;
            n = read_next(argc, argv, &i, "--db", true);
            if (n) c->database_name = n;
        }
        else if(strcmp(argv[i], "--db-tree") == 0)
            c->database_tree = read_next(argc, argv, &i, "--db-tree", false);

        else if(strcmp(argv[i], "--dot") == 0) {
            c->dot_output = true;
            c->dot_file = read_next(argc, argv, &i, "--dot", false);
        }
        else if(strcmp(argv[i], "--") == 0) {
            ++i; /* skip over the -- and look for file names */
            break;
        }
        else if(argv[i][0] == '-') {
            printf("Invalid argument '%s'\n", argv[i]);
            print_help();
            exit(EXIT_FAILURE);
        }
        else
            break;
    }

    if(!force_interface && i < argc) {
        int j;

        c->files = g_new(const char*, argc-i);
        c->numfiles = argc-i;
        for (j = 0; i < argc; ++i, ++j)
            c->files[j] = argv[i];
    }

    if (!c->numfiles)
        c->by_file = false; /* in online mode you can't use by_file */

    if (c->ignore_port && !c->ip_only) {
        fprintf(stderr, "Ignore ports must be used with --ip.\n");
        exit(EXIT_FAILURE);
    }

/* XXX really..? (this is from the old cmd line parsing code, but it's not
   what --help says will happen..)
    if(c->ip_only && !c->noheaders)
        c->noheaders = ETHERNET_HEADER_LENGTH;
*/

    return c;
}

void config_free(config_t *c)
{
    if(c) {
        g_free(c->files);
        g_free(c);
    }
}
