#include"database_version.h"
#include"db.h"
#include"adhic.h"
#include"sim.h"

#include<mysql/mysql.h>
#include<mysql/errmsg.h>
#include<glib.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<unistd.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/fcntl.h>
#include<errno.h>

#define SCAN_MAX_CAPTURES 500

struct db
{
    gint         ref;     /*!< Reference counter */
    MYSQL       *my;      /*!< The main db structure */
    bool         conn;    /*!< Are we connected to the db? */
    char        *host;    /*!< Host where the db is */
    unsigned int port;    /*!< Port to connect to, 0 for default */
    char        *name;    /*!< Name of the database on the server */
    char        *user;    /*!< User name for log in */
    char        *pass;    /*!< Password for log in */
    char        *fifo;    /*!< Path to the FIFO for captures */
    int          fifo_fd; /*!< Reading fd for the captures FIFO */
};

struct db_capture
{
    db_t *db;
    int   id;
    int   node;
    int   seq;
    pcap_dumper_t *dump;
};

struct db_tree {
    unsigned int ref;         /*!< Ref count */
    db_t        *db;
    unsigned int id;          /*!< ID in the database */
    unsigned int current_rev; /*! The current revision ID */
    db_capture_t captures[SCAN_MAX_CAPTURES];
    unsigned int num_captures;
    unsigned int queue_counter; /* The last used queue id */
    unsigned int png_counter;   /* The last used internal node id */
};

#define db_q(db, a, b...) (db_query(db, g_strdup_printf(a, b), false))
#define db_try_q(db, a, b...) (db_query(db, g_strdup_printf(a, b), true))
#define db_id(db, t, format, v...) (db_get_int(db, t, "id", g_strdup_printf(format, v)))
#define db_int(db, t, c, format, v...) (db_get_int(db, t, c, g_strdup_printf(format, v)))
#define db_result(db, t, cols, format, v...) (db_get_result(db, t, cols, g_strdup_printf(format, v)))

static MYSQL_RES* db_query(db_t *db, char *q, bool try);
static void db_create_tree(db_tree_t *t, int rev);
static void db_create_tree_recur(db_tree_t *t, int rev, node_t *n,
                                 unsigned int pos);
static void db_update_tree_traffic(db_tree_t *t, int rev);
static void db_update_tree_traffic_recur(db_tree_t *t, int rev, node_t *n);
static int db_get_int(db_t *db, const char *table, const char *col,
                      char *where);
static MYSQL_RES* db_get_result(db_t *db, const char *table, const char *cols,
                                char *where);
static void db_save_fifo(db_t *db, db_capture_t *cap);

db_t* db_new(const char *host, unsigned int port, const char *dbname,
             const char *user, const char *pass)
{
    char *f;
    db_t *ret;
    MYSQL *my;

    my = mysql_init(NULL);
    if(!my) return NULL;

    ret = g_new0(db_t, 1);
    ret->ref = 1;
    ret->fifo_fd = -1;

    ret->my = my;
    ret->host = g_strdup(host);
    ret->port = port;
    ret->name = g_strdup(dbname);
    ret->user = g_strdup(user);
    ret->pass = g_strdup(pass);

    f = g_strdup_printf("netadict-%s-%05d.fifo", g_get_user_name(),
                        g_random_int_range(0, 100000));
    ret->fifo = g_build_filename(g_get_tmp_dir(), f, NULL);
    g_free(f);

    if(mkfifo(ret->fifo, S_IWUSR | S_IRUSR) == 0) {
        if((ret->fifo_fd = open(ret->fifo, O_RDONLY | O_NDELAY)) == -1) {
            fprintf(stdout,
                    "Unable to open FIFO '%s': %s.\n"
                    "Packet captures will not be possible.\n", ret->fifo,
                    strerror(errno));
            unlink(ret->fifo);
            g_free(ret->fifo);
            ret->fifo = NULL;
        }
    }
    else {
        fprintf(stdout,
                "Unable to make FIFO '%s': %s.\n"
                "Packet captures will not be possible.\n", ret->fifo,
                strerror(errno));
        g_free(ret->fifo);
        ret->fifo = NULL;
    }

    return ret;
}

void db_ref(db_t *db)
{
    ++db->ref;
}

void db_unref(db_t *db)
{
    if(db && --db->ref < 1) {
        if(db->my) {
            mysql_close(db->my);
            if(db->fifo) {
                unlink(db->fifo);
                g_free(db->fifo);
            }
            g_free(db->host);
            g_free(db->name);
            g_free(db->user);
            g_free(db->pass);
        }
        g_free(db);
    }
}

bool db_connect(db_t *db)
{
    MYSQL *ret;

    ret = mysql_real_connect(db->my, db->host, db->user, db->pass, db->name,
                             db->port, NULL, 0);
    if (ret != NULL) {
        MYSQL_RES *res;
        MYSQL_ROW row;
        int ver = 0;

        res = db_try_q(db, "SELECT version FROM schema_info LIMIT 1", 0);

        if(res && (row = mysql_fetch_row(res))) {
            ver = atoi(row[0]);
            mysql_free_result(res);

            if (ver == EXPECTED_DB_VERSION)
                db->conn = TRUE;
        }

        if (!db->conn)
            fprintf(stderr, "Found database schema version %d, but expected "
                    "version %d. Please update your database.\n", ver,
                    EXPECTED_DB_VERSION);
    }

    return db->conn;
}

/*! Run a query on the database. If the connection has been lost, try to
  reestablish it.

  This *frees* the string passed in to it!!
 */
static MYSQL_RES* db_query(db_t *db, char *q, bool try)
{
    MYSQL_RES *res = NULL;

    /*fprintf(stdout, "%s\n", q);*/

    switch (mysql_query(db->my, q)) {
    case 0: /* success, get the results if any */
        res = mysql_store_result(db->my);
        break; 

    case CR_COMMANDS_OUT_OF_SYNC:
        fprintf(stdout, "Illegal query: %s\n", q);
        abort();
        break;

    case CR_SERVER_LOST:
    case CR_SERVER_GONE_ERROR:
        db->conn = false;
        if(!db_connect(db))
            fprintf(stdout, "Connection to database has been lost\n");
        else
            /* dont duplicate q, cuz we're not going to free it inside here */
            return db_query(db, q, try); /* try again, we've reconnected */
        break;
    default:
        if(!try) {
            fprintf(stdout, "Unknown error while executing query: %s\n", q);
            abort();
        }
        break;
    }
    g_free(q);
    return res;
}

/*! The 'where' argument will be freed by this function */
static int db_get_int(db_t *db, const char *table, const char *col,
                      char *where)
{
    MYSQL_RES *res = NULL;
    MYSQL_ROW row;
    int ret = 0;

    res = db_get_result(db, table, col, where);

    if(res && (row = mysql_fetch_row(res))) {
        ret = atoi(row[0]);
        mysql_free_result(res);
    }
    return ret;
}

static MYSQL_RES* db_get_result(db_t *db, const char *table, const char *cols,
                                char *where)
{
    MYSQL_RES *res = NULL;

    res = db_q(db, "SELECT %s FROM %s WHERE %s", cols, table, where);
    g_free(where);
    return res;
}

db_tree_t* db_tree_new(db_t *db, const char *name)
{
    db_tree_t *t = NULL;
    int id;

    g_return_val_if_fail(name != NULL, NULL);
    g_return_val_if_fail(db != NULL, NULL);

    /* find the tree with the given name */
    id = db_id(db, "trees", "tree_name='%s'", name);

    /* delete the tree because we're going to be making a new set of
       revisions for it (this recursively deletes its data) */
    /*if (id) db_q("DELETE FROM trees WHERE tree_name='%s'", name);*/

    if(!id) {
        /* create the new tree */
        db_q(db, "INSERT INTO trees SET tree_name='%s'", name);
        id = db_id(db, "trees", "tree_name='%s'", name);
    }

    if(id) {
        t = g_new0(db_tree_t, 1);
        t->ref = 1;
        t->db = db;
        db_ref(db);
        t->id = id;

        /* get the latest revision if there is one */
        t->current_rev = db_id(db, "revisions",
                               "tree_id=%d ORDER BY revision DESC LIMIT 1",
                               id);
        t->queue_counter = db_int(db, "revisions", "MAX(queue_counter)",
                                  "tree_id=%d GROUP BY tree_id", id);
        t->png_counter = db_int(db, "revisions", "MAX(png_counter)",
                                "tree_id=%d GROUP BY tree_id", id);
    }
    return t;
}

void db_tree_ref(db_tree_t *t)
{
    if(!t) return;
    t->ref++;
}

void db_tree_unref(db_tree_t *t)
{
    if(t && --t->ref == 0) {
        unsigned int i;

        for(i = 0; i < t->num_captures; ++i) {
            pcap_dump_close(t->captures[i].dump);
            t->captures[i].id = 0;
            t->captures[i].node = 0;
            t->captures[i].dump = NULL;
        }

        db_unref(t->db);
        g_free(t);
    }
}

bool db_tree_get_current_counters(db_tree_t *t, unsigned int *queue,
                                  unsigned int *png)
{
    if(t->current_rev) {
        *queue = t->queue_counter;
        *png = t->png_counter;
        return true;
    }
    else
        return false;
}

void db_create_new_tick(db_tree_t *t, const time_t time,
                        unsigned int queue_counter, unsigned int png_counter)
{
    struct tm *tm;
    int rev;

    g_return_if_fail(t != NULL);

    if(!(tm = gmtime(&time))) {
        fprintf(stdout,
                "ERROR: Invalid date passed to db_create_new_tick: %ld\n",
                (long) time);
        abort();
    }

    /* overwrite the revision and anything past it */
    db_q(t->db, "DELETE FROM revisions WHERE tree_id=%d AND "
         "revision>='%d-%d-%d %d:%d:%d'",
         t->id, tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
         tm->tm_hour, tm->tm_min, tm->tm_sec);

    /* create the new revision */
    db_q(t->db, "INSERT INTO revisions SET tree_id=%d, "
         "revision='%d-%d-%d %d:%d:%d', queue_counter=%d, png_counter=%d",
         t->id, tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
         tm->tm_hour, tm->tm_min, tm->tm_sec,
         queue_counter, png_counter);

    /* get the revision id */
    rev = db_id(t->db, "revisions",
                "tree_id=%d AND revision='%d-%d-%d %d:%d:%d'",
                t->id, tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
                tm->tm_hour, tm->tm_min, tm->tm_sec);

    db_create_tree(t, rev);

    db_update_tree_traffic(t, rev);
    t->current_rev = rev;
}

static void db_create_tree_recur(db_tree_t *t, int rev, node_t *n,
                                 unsigned int pos)
{
    const char *table;
    int id;

    assert(n->type == PNG_OP || n->type == QUEUE);

    if(n->type == PNG_OP)
        table = "internal_nodes";
    else
        table = "terminal_nodes";

    /* create the node */
    db_q(t->db, "INSERT INTO %s SET "
         "node_id=%d, "
         "revision_id=%d, "
         "position=%d",
         table, n->id, rev, pos);
    id = db_id(t->db, table, "node_id=%d AND revision_id=%d AND position=%d",
               n->id, rev, pos);

    if(n->type == PNG_OP) {
        /* internal node */
        int i;
        char *value;

        value = g_new0(char, N*2+1);

        /* create the pngrams */
        for(i = 0; i < n->data.pngs.png_size; ++i) {
            int png_id;

            /* get the png-value */
            mysql_real_escape_string(t->db->my, value,
                                     (char*)n->data.pngs.pngs[i]->val,
#ifdef VARIABLE_N
                                     n->data.pngs.pngs[i]->length
#else
                                     N
#endif
                );

            /* put it in the database if it's not there yet */
            png_id = db_id(t->db, "pngrams", "p=%d AND n='%s'",
                           n->data.pngs.pngs[i]->offset, value);
            if(!png_id) {
                db_q(t->db, "INSERT INTO pngrams SET p=%d, n='%s'",
                     n->data.pngs.pngs[i]->offset, value);
                png_id = db_id(t->db, "pngrams", "p=%d AND n='%s'",
                               n->data.pngs.pngs[i]->offset, value);
            }

            /* connect the node to the p,n-gram */
            db_q(t->db, "INSERT INTO node_pngrams SET "
                 "internal_node_id=%d, "
                 "pngram_id=%d",
                 id, png_id);
        }
        g_free(value);

        db_create_tree_recur(t, rev, n->left, pos * 2);
        db_create_tree_recur(t, rev, n->right, pos * 2 + 1);
    } else {
        /* terminal node */

        /* create the traffic row for the node */
        db_q(t->db, "INSERT INTO traffic SET "
             "terminal_node_id=%d", id);
    }
}

static void db_create_tree(db_tree_t *t, int rev)
{
    db_create_tree_recur(t, rev, node_head(), 1);
}

static void db_update_tree_traffic_recur(db_tree_t *t, int rev, node_t *n)
{
    const char *table;

    assert(n->type == PNG_OP || n->type == QUEUE);

    if(n->type == PNG_OP) {
        /* internal node */
        db_update_tree_traffic_recur(t, rev, n->left);
        db_update_tree_traffic_recur(t, rev, n->right);
    } else {
        p_stats totals;
        int i, id;
        GString *q;
        bool first;

        id = db_id(t->db, "terminal_nodes",
                   "revision_id=%d AND node_id=%d",
                   rev, n->id);
        assert(id);

        p_stats_names *names = p_stats_get_db_names();

        p_stats_init(&totals);
        /* XXX is this 1 tick of data? */
        p_stats_total(&totals, &(n->data.queue.stats[simtime % MAX_TIME]), 1);

        /* write the traffic */
        q = g_string_new("UPDATE traffic SET");
        first = true;

        for(i = 0; i < ETHERTYPE_ENUM_SIZE; ++i)
            if(names->ether_names[i]) {
                if(!first) g_string_append(q, ",");
                else first = false;
                g_string_append_printf(q, " %s=%d",
                                       names->ether_names[i],
                                       totals.ether_stats[i]);
            }
        for(i = 0; i < IPTYPE_SIZE; ++i)
            if(names->ip_names[i]) {
                if(!first) g_string_append(q, ",");
                else first = false;
                g_string_append_printf(q, " %s=%d",
                                       names->ip_names[i],
                                       totals.ip_stats[i]);
            }
        for(i = 0; i < UDPTYPE_SIZE; ++i)
            if(names->udp_names[i]) {
                if(!first) g_string_append(q, ",");
                else first = false;
                g_string_append_printf(q, " %s=%d",
                                       names->udp_names[i],
                                       totals.udp_stats[i]);
            }
        for(i = 0; i < TCPTYPE_SIZE; ++i)
            if(names->tcp_names[i]) {
                if(!first) g_string_append(q, ",");
                else first = false;
                g_string_append_printf(q, " %s=%d",
                                       names->tcp_names[i],
                                       totals.tcp_stats[i]);
            }

        g_string_append_printf(q, " WHERE "
                               "terminal_node_id=%d",
                               id);
        db_query(t->db, q->str, false);
        g_string_free(q, false); /* the insides have already been freed */
    }
}

static void db_update_tree_traffic(db_tree_t *t, int rev)
{
    db_update_tree_traffic_recur(t, rev, node_head());
}

bool db_tree_foreach_node(db_tree_t *t, db_foreach_node_func func,
                          void *data)
{
    MYSQL_RES *res;
    MYSQL_ROW row;

    /* is there is a revision to get nodes for? */
    if(!t->current_rev) return false;

    /* get each internal node (which has 1 or more p,n-grams with it) */
    res = db_result(t->db, "internal_nodes", "id,node_id,position",
                    "revision_id=%d ORDER BY position", t->current_rev);
    while(res && (row = mysql_fetch_row(res))) {
        MYSQL_RES *pngres;
        MYSQL_ROW pngrow;
        db_png_t *pngs = NULL;
        gint n_pngs = 0;

        /* get all of its p,n-grams */
        pngres = db_q(t->db, "SELECT p,n FROM node_pngrams "
                      "INNER JOIN pngrams ON pngram_id=pngrams.id "
                      "WHERE internal_node_id=%d", atoi(row[0]));
        while(pngres && (pngrow = mysql_fetch_row(pngres))) {
            unsigned long len = mysql_fetch_lengths(pngres)[1];

            pngs = g_renew(db_png_t, pngs, n_pngs+1);
            pngs[n_pngs].p = atoi(pngrow[0]);
            pngs[n_pngs].len = len;
            pngs[n_pngs].n = g_memdup(pngrow[1], len);
            ++n_pngs;
        }
        if(pngres) mysql_free_result(pngres);

        /* call the callback */
        func(PNG_OP, atoi(row[1]), atoi(row[2]), pngs, n_pngs, data);
    }
    if(res) mysql_free_result(res);

    /* get the terminal nodes */
    res = db_result(t->db, "terminal_nodes", "node_id,position",
                    "revision_id=%d ORDER BY position", t->current_rev);
    while(res && (row = mysql_fetch_row(res))) {
        /* call the callback */
        func(QUEUE, atoi(row[0]), atoi(row[1]), NULL, 0, data);
    }
    if(res) mysql_free_result(res);

    return true;
}

void db_tree_lookup_capture_requests(db_tree_t *t, pcap_t *pcap)
{
    MYSQL_RES *res;
    MYSQL_ROW row;
    unsigned int i, j, n;
    db_capture_t new_captures[SCAN_MAX_CAPTURES];

    /* find what nodes the user wants captures for */
    res = db_result(t->db, "capture_requests", "id,node_id",
                    "tree_id=%d", t->id);
    n = 0;
    while(res && (row = mysql_fetch_row(res)) && n < SCAN_MAX_CAPTURES) {
        new_captures[n].id = atoi(row[0]);
        new_captures[n].node = atoi(row[1]);
        ++n;
    }
    if(res) mysql_free_result(res);

    /* copy stuff over when the request hasn't changed */
    for(i = 0; i < n; ++i) {
        bool found = false;
        /* is this id from before? */
        for(j = 0; j < t->num_captures; ++j)
            if(new_captures[i].id == t->captures[j].id) {
                new_captures[i].db = t->captures[j].db;
                new_captures[i].seq = t->captures[j].seq;
                new_captures[i].dump = t->captures[j].dump;
                found = true;
                break;
            }
        if(!found) {
            new_captures[i].db = t->db;
            new_captures[i].seq = 0;
            if(t->db->fifo) {
                new_captures[i].dump = pcap_dump_open(pcap, t->db->fifo);
                db_save_fifo(t->db, &new_captures[i]);
            }
        }
    }

    /* close captures that have ended */
    for(i = 0; i < t->num_captures; ++i) {
        bool found = false;
        for(j = 0; j < n; ++j)
            if(t->captures[i].id == new_captures[j].id) {
                found = true;
                break;
            }
        if(!found) {
            pcap_dump_close(t->captures[i].dump);
            t->captures[i].dump = NULL;
        }
    }

    /* copy the new stuff into our data structures */
    t->num_captures = n;
    for(i = 0; i < n; ++i)
        t->captures[i] = new_captures[i];
}

int db_tree_num_capture_requests(db_tree_t *t)
{
    return t->num_captures;
}

db_capture_t* db_tree_capture_request(db_tree_t *t, unsigned int i)
{
    assert(i < t->num_captures);
    return &t->captures[i];
}

void db_capture_packet(db_capture_t *cap, const struct pcap_pkthdr *h,
                       const u_char *sp)
{
    if(cap->dump) {
        /* dump to the fifo.. */
        pcap_dump((u_char*) cap->dump, h, sp);
        /* then read it back.. */
        db_save_fifo(cap->db, cap);
    }
}

static void db_save_fifo(db_t *db, db_capture_t *cap)
{
#define READ_BLOCK 1024
    char read_buf[READ_BLOCK];
    char escaped[READ_BLOCK * 2 + 1];
    ssize_t r;

    if(!db->fifo) return;

    pcap_dump_flush(cap->dump);
    do {
        r = read(db->fifo_fd, read_buf, READ_BLOCK);
        if(r > 0) {
            mysql_real_escape_string(db->my, escaped,
                                     read_buf, (unsigned long) r);

            /* this may fail if the request is no longer present in the
               database */
            db_try_q(db, "INSERT INTO captures SET "
                     "capture_id=%d, sequence=%d, data='%s'",
                     cap->id, cap->seq, escaped);
            ++cap->seq;
        }
        else if(r < 0) {
            if(errno == EAGAIN) {
                r = 0; /* no data to read yet */
            }
            else if(errno == EINTR) {
                ;/* try again for EINTR */
            }
            else {
                fprintf(stdout, "Error reading the capture FIFO: %s\n",
                        strerror(errno));
                abort();
            }
        }
    } while(r != 0);
}

int db_capture_node_id(db_capture_t *cap)
{
    return cap->node;
}
