require 'rubygems'
gem 'libxml-ruby'
require 'xml/libxml'

module SVG

class Color
  def initialize(r, g=0, b=0)
    if r.kind_of?(String)
      @colorstring = r
    else      
      r = (r * 255).to_i if r.kind_of?(Float)
      g = (g * 255).to_i if g.kind_of?(Float)
      b = (b * 255).to_i if b.kind_of?(Float)
      r, g, b = r.to_i, g.to_i, b.to_i
      if (r < 0 || r > 255 || g < 0 || g > 255 || b < 0 || b > 255)
        raise 'Invalid color, values out of range'
      end
      @colorstring = "#%02x%02x%02x" % [r, g, b]
    end
  end

  def to_s
    @colorstring
  end
end

class Point
  attr :x
  attr :y

  def initialize(x, y)
    @x, @y = x, y
  end
end

class Text
  attr :text, true
  attr :color, true
  def initialize(text)
    @text = text
    @color = Color.new(0, 0, 0) # black
  end

  def to_xml(width, height)
    text = XML::Node.new('text')
    text['font-family'] = 'sans-serif'
    text['font-weight'] = 'normal'
    text['font-style'] = 'normal'
    text['x'] = '50%'
    text['y'] = '50%'
    text['stroke'] = @color.to_s
    text['fill'] = @color.to_s
    text['text-anchor'] = 'middle'
    text['fill-opacity'] = 1.0.to_s

    lines = @text.split("\n")

    # how big should the text be?
    size = height / lines.length
    size = (size < 7 ? 7 : (size > 30 ? 30 : size)) # between 7 and 30
    text['font-size'] = "#{size}px"

    first = true
    lines.each {|i|
      line = XML::Node.new('tspan')
      line['x'] = '50%'
      if (first)
        line['y'] = ((height - size * lines.length + size) / 2).to_s
        first = false
      else
        line['dy'] = size.to_s
      end
      line.content = i
      text.child = line
    }

    text
  end
end

class Line
  attr :color, true
  attr :width, true
  attr :x1, true
  attr :y1, true
  attr :x2, true
  attr :y2, true

  def initialize(id)
    @width = 1
    @color = Color.new(0, 0, 0) #black
    @from = Point.new(0, 0)
    @to = Point.new(10, 10)
  end

  def to_xml()
    raise 'Invalid width: ' + @width.to_s if (@width <= 0)

    parent = XML::Node.new('line');
    # parent['id'] = @id
    parent['x1'] = @x1.to_s
    parent['y1'] = @y1.to_s
    parent['x2'] = @x2.to_s
    parent['y2'] = @y2.to_s
    parent['stroke'] = @color.to_s
    parent['stroke-width'] = @width.to_s

    parent
  end
end

class Node
  attr :xradius
  attr :yradius
  attr :bordercolor, true
  attr :borderwidth
  attr :tooltip, true
  attr :text, true
  attr :x, true
  attr :y, true

  def initialize(id, xradius, yradius)
    @id = id.to_s
    @xradius, @yradius = xradius, yradius
    @borderwidth = 1
    @bordercolor = Color.new(0, 0, 0) #black
    @tooltip = nil
    @text = nil
    @x = @y = 0
  end

  def width()
    width = @xradius * 2 + @borderwidth * 2;
  end
  def height()
    height = @yradius * 2 + @borderwidth * 2;
  end

  def to_xml()
    raise 'Invalid radius: ' + @xradius.to_s if (@xradius <= 0)
    raise 'Invalid radius: ' + @yradius.to_s if (@yradius <= 0)

    width = self.width
    height = self.height

    parent = XML::Node.new('svg');
    parent['id'] = @id
    parent['x'] = @x.to_s
    parent['y'] = @y.to_s
    parent['width'] = width.to_s
    parent['height'] = height.to_s
    parent['viewBox'] = "0 0 #{width} #{height}"

    # make the path node
    ell = XML::Node.new('ellipse')
    ell['cx'] = (width/2).to_s
    ell['cy'] = (height/2).to_s
    ell['rx'] = @xradius.to_s
    ell['ry'] = @yradius.to_s

    ell['fill-opacity'] = 0.to_s
    ell['stroke'] = @bordercolor.to_s
    ell['stroke-width'] = @borderwidth.to_s
    if (@tooltip)
      ell['tooltip'] = 'enable'
      ell['hint'] = @tooltip.to_s
    end

    #  draw text if there is
    if (@text != nil) then parent.child = @text.to_xml(width, height) end

    parent.child = ell
    parent
  end
end

class Pie
  # bitflags..
  RING_NONE  = 0
  RING_IP    = 1
  RING_TCPIP = 2
  RING_UDPIP = 3
  RING_NUM_RINGS = 4

  class Slice
    attr :ring, true
    attr :percent, true
    attr :color, true
    attr :tooltip, true
    def initialize(ring, percent, color, tooltip)
      @ring, @percent, @color, @tooltip = ring, percent, color, tooltip
    end
  end

  class Ring
    attr :color, true
    attr :tooltip, true
    def initialize()
      @color = Color.new(1.0, 1.0, 1.0)
      @tooltip = nil
    end
  end

  attr :radius
  attr :bordercolor, true
  attr :borderwidth
  attr :slices, true
  attr :boxcolor, true
  attr :ringwidth
  attr :rings
  attr :text, true
  attr :x, true
  attr :y, true

  def initialize(id, radius)
    @id = id.to_s
    @slices = []
    @bordercolor = Color.new(0,0,0) # black borders
    @borderwidth = 1
    @autoringwidth = true
    @boxcolor = Color.new(0.6, 0.6, 0.6) # grey background-box
    @text = nil
    @x = @y = 0

    self.radius = radius # use the attribute setting function

    @rings = []
    RING_NUM_RINGS.times { |i| @rings[i] = Ring.new }
  end

  def width()
    @size
  end
  def height()
    @size
  end

  def ringwidth=(w)
    @autoringwidth = false
    @ringwidth = w
  end

  def radius=(radius)
    @radius = radius
    @size = @radius * 2 + @borderwidth * 2;
    if (@autoringwidth)
      @ringwidth = radius / 20 < 1 ? 1 : radius / 12
    end
  end

  def borderwidth=(bw)
    @borderwidth = bw
    @size = @radius * 2 + @borderwidth * 2;
  end

  def add_slice(ring, percent, color, tooltip = nil)
    raise "Invalid ring selection" if (ring < 0 || ring >= RING_NUM_RINGS)
    s = Slice.new(ring, percent, color, tooltip)
    @slices.push s
    s
  end

  def to_xml()
    if (@radius <= 0) then raise 'Invalid radius: ' + @radius.to_s end

    # validate the slices
    sum = 0;
    @slices.each { |s| sum += s.percent }
    sum = sum.to_i
    raise 'Pie slices add up to ' + sum.to_s + '%' if sum != 100 && sum != 0

    parent = XML::Node.new('svg');
    parent['id'] = @id
    parent['x'] = @x.to_s
    parent['y'] = @y.to_s
    parent['width'] = @size.to_s
    parent['height'] = @size.to_s
    parent['viewBox'] = "0 0 #{@size} #{@size}"

    # draw a box around it, if it has only a single slice, or no slices
    if (@slices.length <= 1)
      parent.child = render_box
    end

    if (@slices.length > 0)
      # group the slices by ring
      slicesbyring = []
      RING_NUM_RINGS.times { slicesbyring.push [] }
      @slices.each { |s| slicesbyring[s.ring].push s }

      start_angle = 0.0
      group_start_angle = []
      group_end_angle = []

      # render the slices by ring grouping
      RING_NUM_RINGS.times { |i|
        group_start_angle[i] = start_angle

        slicesbyring[i].each { |s|
          num_rings = case i
                      when RING_NONE  then 0
                      when RING_IP    then 1
                      when RING_TCPIP then 2
                      when RING_UDPIP then 2
                      else raise "Invalid ring"
                      end

          end_angle = start_angle + Math::PI * s.percent / 50.0

          parent.child = render_slice(s, 0, @radius, start_angle, end_angle)
          start_angle = end_angle # for the next slice of the pie
        }

        group_end_angle[i] = start_angle
      }

      # render the IP ring
      start_angle = group_start_angle[RING_IP]
      end_angle = group_end_angle[RING_UDPIP]
      inner_radius = @radius - @ringwidth
      outer_radius = @radius
      if (end_angle > start_angle)
        parent.child = render_slice(@rings[RING_IP],
                                    inner_radius, outer_radius,
                                    start_angle, end_angle)
      end

      # render the TCP ring
      start_angle = group_start_angle[RING_TCPIP]
      end_angle = group_end_angle[RING_TCPIP]
      inner_radius = @radius - @ringwidth * 2
      outer_radius = @radius - @ringwidth
      if (end_angle > start_angle)
        parent.child = render_slice(@rings[RING_TCPIP],
                                    inner_radius, outer_radius,
                                    start_angle, end_angle)
      end

      # render the UDP ring
      start_angle = group_start_angle[RING_UDPIP]
      end_angle = group_end_angle[RING_UDPIP]
      inner_radius = @radius - @ringwidth * 2
      outer_radius = @radius - @ringwidth
      if (end_angle > start_angle)
        parent.child = render_slice(@rings[RING_UDPIP],
                                    inner_radius, outer_radius,
                                    start_angle, end_angle)
      end
    end

    #  draw text if there is
    if (@text != nil) then parent.child = @text.to_xml(@size, @size) end

    parent
  end

  private
  def render_slice(slice, inner_radius, radius, start_angle, end_angle)
    center = @size / 2
    mid_angle = start_angle + (end_angle - start_angle) / 2

    inner_end_x = center - (inner_radius * Math::sin(start_angle))
    inner_end_y = center - (inner_radius * Math::cos(start_angle))
    inner_mid_x = center - (inner_radius * Math::sin(mid_angle))
    inner_mid_y = center - (inner_radius * Math::cos(mid_angle))

    start_x = center - (radius * Math::sin(start_angle))
    start_y = center - (radius * Math::cos(start_angle))
    mid_x = center - (radius * Math::sin(mid_angle))
    mid_y = center - (radius * Math::cos(mid_angle))

    end_angle -= 2 * Math::PI if end_angle >= 2 * Math::PI

    inner_start_x = center - (inner_radius * Math::sin(end_angle))
    inner_start_y = center - (inner_radius * Math::cos(end_angle))

    end_x = center - (radius * Math::sin(end_angle))
    end_y = center - (radius * Math::cos(end_angle))

    # make the path node
    path = XML::Node.new('path')
    path['d'] = ''

    if (inner_radius > 0)
      # move to the start of the inner arc
      path['d'] += "M#{inner_start_x},#{inner_start_y} "
      # draw the inner arc
      path['d'] +=
        "A#{inner_radius},#{inner_radius} 0 0,1" +
        " #{inner_mid_x},#{inner_mid_y} " +
        "A#{inner_radius},#{inner_radius} 0 0,1 #{inner_end_x},#{inner_end_y} "
    else
      # move to the end of the inner arc
      path['d'] += "M#{inner_end_x},#{inner_end_y} "
    end

    # move to the start of the outer arc
    path['d'] += "L#{start_x},#{start_y} "

    # draw the outer arc
    path['d'] +=
      "A#{radius},#{radius} 0 0,0 #{mid_x},#{mid_y} " +
      "A#{radius},#{radius} 0 0,0 #{end_x},#{end_y} "

    # close the path
    path['d'] += "Z"

    path['fill'] = slice.color.to_s
    path['stroke'] = @bordercolor.to_s
    path['stroke-width'] = @borderwidth.to_s
    if (slice.tooltip)
      path['tooltip'] = 'enable'
      path['hint'] = slice.tooltip.to_s
    end

    path
  end

  def render_box
    # make the path node
    path = XML::Node.new('path')

    left = top = @size / 2 - @radius
    right = bottom = @size / 2 + @radius
    corner = @radius / 5

    path['d'] = "M#{right-corner},#{top} " +
      "A#{corner},#{corner} 0 0,1 #{right},#{top+corner} " +
      "L#{right},#{bottom-corner} " +
      "A#{corner},#{corner} 0 0,1 #{right-corner},#{bottom} " +
      "L#{left+corner},#{bottom} " +
      "A#{corner},#{corner} 0 0,1 #{left},#{bottom-corner} " +
      "L#{left},#{top+corner} " +
      "A#{corner},#{corner} 0 0,1 #{left+corner},#{top} Z"
    path['fill'] = @boxcolor.to_s
    path['stroke'] = @bordercolor.to_s
    path['stroke-width'] = @borderwidth.to_s

    path
  end
end

class ColorKey
  attr :color, true
  attr :bordercolor, true
  attr :borderwidth, true
  attr :tooltip, true
  attr :size, true

  def initialize(size, color)
    @color = color
    @bordercolor = Color.new(0, 0, 0) #black
    @borderwidth = 1
    @tooltip = nil
    @size = size
  end

  def to_xml()
    s = @size - @borderwidth * 2

    raise 'Invalid size: ' + s.to_s if (s <= 0)

    n = XML::Node.new('rect');
    n['x'] = @borderwidth.to_s
    n['y'] = @borderwidth.to_s
    n['width'] = s.to_s
    n['height'] = s.to_s
    n['fill'] = @color.to_s
    n['stroke'] = @bordercolor.to_s
    n['stroke-width'] = @borderwidth.to_s
    if (@tooltip)
      n['tooltip'] = 'enable'
      n['hint'] = @tooltip.to_s
    end

    n
  end
end

end
